import 'dart:async';
import 'package:analyzer/dart/element/element.dart';
import 'package:recase/recase.dart';
import 'package:source_gen/source_gen.dart';

import 'package:rest_api_builder/rest_api_builder.dart';

String _generateParamsCode(Set<String> params, Set<String> queries, String req, bool isVoidRequest) {
  String ret = '';
  if (params.isNotEmpty) {
    ret += ', ' + params.map((param) => 'String $param').join(', ');
  }
  if (!isVoidRequest || queries.isNotEmpty) {
    ret += ', {';
    if (!isVoidRequest) {
      ret += '$req request';
    }
    if (queries.isNotEmpty) {
      if (!isVoidRequest) {
        ret += ', ';
      }
      ret += queries.map((param) => '${ReCase(param).camelCase}').join(', ');
    }
    ret += '}';
  }
  return ret;
}

String _generateReturnResponse(String res, String resEncoding) {
  switch (res) {
    case 'String':
      return 'response.body';
    case 'void':
      return '';
    default:
      return '$res.fromJson(json.decode(${_convertResponseBodyForEncoding(resEncoding)}))';
  }
}

String _convertResponseBodyForEncoding(String encoding) {
  if (encoding == null) {
    return 'response.body';
  }
  // TODO(wonjerry): Add other encodingType
  switch (encoding) {
    case 'utf-8':
      return 'utf8.decode(response.bodyBytes)';
    default:
      throw Exception('rest_api_builder_generator - not supported encoding: $encoding.\nSupported encodings: utf-8');
  }
}

FutureOr<String> _generateCode(String method, Element element, ConstantReader annotation) {
  final String className = element.displayName;
  String path = annotation.read('path').stringValue;
  // TODO(sglim): Handle path required String. like '/account/{int uid}/

  final String req = annotation.read('req').stringValue;
  final String res = annotation.read('res').stringValue;
  final isResEncodingNull = annotation.read('resEncoding').isNull;
  final resEncoding = isResEncodingNull ? null : annotation.read('resEncoding').stringValue;

  final bool isVoidRequest = req == 'void';
  final String bodyString = isVoidRequest ? '' : 'body: json.encode(request.toJson()),';
  final String bodyDebugPrint =
      isVoidRequest ? '' : '\\n\${const JsonEncoder.withIndent(\'  \').convert(request.toJson())}';

  final RegExp exp = RegExp(r'{(\w+)}');
  final Iterable<Match> matches = exp.allMatches(path);
  // Group 0 is {abc}, Group1 is abc.
  final Set<String> params = matches.map((match) => match.group(1)).toSet();
  for (final param in params) {
    path = path.replaceAll('{$param}', '\$$param');
  }

  // Url
  String urlDefinition = 'final url = \'\${client.host}$path\';';
  String urlCode = '\$url';

  // Query
  Set<String> queries = {};
  final queryAnnotation = annotation.read('query');
  final hasQuery = !queryAnnotation.isNull && queryAnnotation.isList;
  if (hasQuery) {
    queries = queryAnnotation.listValue.map((dartObject) => dartObject.toStringValue()).toSet();
    const queryStringConverter = 'final _ = (k, v) => v != null ? \'\$k=\$v\' : \'\';';
    final queryString = queries.map((q) => '\$\{_(\'$q\', ${ReCase(q).camelCase})}').join('\' \'&');
    urlDefinition += '\n$queryStringConverter\nfinal query = \'?$queryString\';\n';
    urlCode += '\$query';
  }

  final paramsCode = _generateParamsCode(params, queries, req, isVoidRequest);

  return '''
Future<$res> _\$$method$className(client$paramsCode) async {
  ${urlDefinition}print('[$method] $urlCode$bodyDebugPrint');
  final response = await client.$method('$urlCode',
    $bodyString
  );

  if (HttpStatus.isSuccess(response.statusCode)) {
    return ${_generateReturnResponse(res, resEncoding)};
  } else {
    throw ApiException(
      method: '$className',
      statusCode: response.statusCode,
      reasonPhrase: response.reasonPhrase,
      body: response.body,
    );
  }
  // Do we need to close the connection after a request?
}
  ''';
}

class GetBuilderGenerator extends GeneratorForAnnotation<Get> {
  @override
  FutureOr<String> generateForAnnotatedElement(Element element, ConstantReader annotation, buildStep) {
    // TODO(sglim): Handle Get path param.
    return _generateCode('get', element, annotation);
  }
}

class PostBuilderGenerator extends GeneratorForAnnotation<Post> {
  @override
  FutureOr<String> generateForAnnotatedElement(Element element, ConstantReader annotation, buildStep) {
    return _generateCode('post', element, annotation);
  }
}

class PutBuilderGenerator extends GeneratorForAnnotation<Put> {
  @override
  FutureOr<String> generateForAnnotatedElement(Element element, ConstantReader annotation, buildStep) {
    return _generateCode('put', element, annotation);
  }
}

class DeleteBuilderGenerator extends GeneratorForAnnotation<Delete> {
  @override
  FutureOr<String> generateForAnnotatedElement(Element element, ConstantReader annotation, buildStep) {
    return _generateCode('delete', element, annotation);
  }
}
